/*
 * 	timewarp v0.1, (c) 1988 Jeremy D. Impson <jdimpson@acm.org>
 *
 *      Based on "herion.c", written by Runar Jensen <zarq@opaque.org>
 *	and on "totem" by lilo by way of my other creation, "paudio", 
 *	and the Linux kernel sources
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Compile with:
 * gcc -O2 -fomit-frame-pointer -DVERSION=\"X.Y\" -c timewarp.c
 *
 */

/*
 * BUGS:
 *	While multiple reads from /proc/warp can be done, any entries
 *	written _MUST_ be done in one write system call.
 *
 *	While superuser can see all entries, it can only remove from
 *	its own entry.
 *
 *	This is the most complicated thing I've done in kernel space,
 *	where I already don't know what I'm doing, so be careful...
 */

/*
 *      Only works in kernel mode: force on the macros
 */

#ifndef MODULE
#  define MODULE
#endif
#ifndef __KERNEL__
#  define __KERNEL__
#endif

/* #define STAY_DEBUG */
#define COMMLENGTH 16 /* this should be in the kernel somwhere! */

#include <linux/fs.h>
#include <linux/module.h>
#include <linux/modversions.h>
#include <linux/malloc.h>
#include <linux/unistd.h>
#include <sys/syscall.h>
#include <linux/timex.h>
#include <linux/proc_fs.h>
#include <asm/semaphore.h>
#include <stdlib.h>

/*
 * datatypes
 */

/* structure to store one line of an entry */
struct line {
        char progname[COMMLENGTH];       /* program to fool -- 16 because this is the size in the task_struct def */
        int offset;             /* how much to fool it by (subtract from current current time) */ 
	struct line * next;     /* next line */
};

/* structure that represents one uid's entry */
struct entry {
        uid_t uid;                      /* who this entry is for */
        struct line * firstline;        /* first line in this entry */
        struct entry * next;            /* next entry */
	struct semaphore sem;		/* must be MUTEX */
};

/*
 * global variables
 */

struct entry * entryhead;	/* init'd in init_module */
int errno;
static int timewarp_indodeval = 0;
extern void *sys_call_table[];
extern struct timezone sys_tz;


/*
 * function declarations
 */

static inline _syscall2(int, gettimeofday, struct timeval *, tv, struct timezone *, tz); 
static inline _syscall1(int, time, int *, tloc);

int (*original_gettimeofday)(struct timeval *tv, struct timezone *tz);
int (*original_time)(int * tloc);

int	init_module(void);
void	cleanup_module(void);
int	sucker();
int	findoffset (struct line * lineptr);
int	hacked_gettimeofday (struct timeval *tv, struct timezone *tz);
int	hacked_time (int * tloc);
static	int timewarp_read (struct inode * inode, struct file * file,
			char * buf,       int count);
static	int timewarp_get_info (char *buf, char **start, off_t offset, 
			int len, int unused);
static	int timewarp_write(struct inode * inode , struct file * file, 
			const char * buf, int len);

/*
 *      file operations for /proc/warp
 */

struct file_operations timewarp_operations = {
	NULL,			/* lseek       */
	timewarp_read,		/* read        */
	timewarp_write,		/* write       */
	NULL,			/* readdir     */
	NULL,			/* select      */
	NULL,			/* ioctl       */
	NULL,			/* mmap        */
	NULL,			/* open        */
	NULL,			/* close       */
	NULL			/* can't fsync */
};

/*
 *      inode operations for /proc/warp
 */

struct inode_operations timewarp_inode_operations = {
        &timewarp_operations,		/* file-ops    */
        NULL,				/* create      */
        NULL,				/* lookup      */
        NULL,				/* link        */
        NULL,				/* unlink      */
        NULL,				/* symlink     */
        NULL,				/* mkdir       */
        NULL,				/* rmdir       */
        NULL,				/* mknod       */
        NULL,				/* rename      */
        NULL,				/* readlink    */
        NULL,				/* follow_link */
        NULL,				/* readpage    */
        NULL,				/* writepage   */
        NULL,				/* bmap        */
        NULL,				/* truncate    */
        NULL 				/* permission  */
};

/*
 *      Directory entry for /proc/warp, read-write.
 */

struct proc_dir_entry timewarp_dir =
{
        0,				/* low_ino    */
        4,				/* namelen    */
        "warp",				/* name       */
        S_IFREG | S_IRUGO |S_IWUGO,	/* mode       */
        1,				/* nlink      */
        0,				/* uid        */
        0,				/* gid        */
        0,				/* size       */
        &timewarp_inode_operations,         /* ops        */
        timewarp_get_info,			/* get_info   */
	NULL				/* fill_inode */
};

/*  string functions */

int myatoi(char *str)
{
	int res = 0;
	int mul = 1;
	char *ptr;

	for(ptr = str + strlen(str) - 1; ptr >= str; ptr--) {
		if(*ptr < '0' || *ptr > '9')
			return(-1);
		res += (*ptr - '0') * mul;
		mul *= 10;
	}
	return(res);
}

inline void mystrcpy(char *src, char *dst) {

	while(*src)
		*(dst++) = *(src++);
	*(dst) = '\0'; 

}

inline void mystrncpy(char *src, char *dst, int l) {
	int i = 0;

	while (*src && i < l) {
		*(dst++) = *(src++);
		i++;
	}
	*(dst) = '\0'; /* XXX: uhh, this might be a bug... */
}
int mystrcmp(char *str1, char *str2)
{
	while(*str1 && *str2)
		if(*(str1++) != *(str2++))
			return(-1);
	return(0);
}

int mystrncmp(char *str1, char *str2, int l) {
	int i = 0;
	while(*str1 && *str2 && i < l) {
		if(*(str1++) != *(str2++))
			return(-1);
		i++;
	}
	return(0);
}


/*
 * various helper functions, mostly for looking things up in the struct entry
 * and struct line data structures.
 */

/* find the current proccesses (uid's) entry; return NULL if it doesn't exist */
struct entry * find_entry() {

        struct entry * entry;

        for (entry = entryhead; entry; entry = entry->next)
                if (entry->uid == current->uid)
                        return (entry);

        return NULL;
}

/* make an entry */
inline struct entry * make_entry() {

	struct entry * entry = NULL;

	entry = kmalloc(sizeof(struct entry), GFP_KERNEL);
	if (!entry) return (NULL);
	entry->uid = current->uid;
	entry->firstline = NULL;
	entry->sem = MUTEX;

	/* returns <0 if we were interrupted */
#ifdef STAY_DEBUG
	printk("down on %i (entryhead) sem in make_entry()\n", entryhead->uid);
#endif
	if (down_interruptible(&(entryhead->sem)) != 0) return (NULL);
#ifdef STAY_DEBUG
	printk("down on %i sem in make_entry()\n", entry->uid);
#endif
	down_interruptible(&(entry->sem)); /* the next entryhead */
	entry->next = entryhead;
	entryhead   = entry;
#ifdef STAY_DEBUG
	printk("up on %i (old entryhead) sem in make_entry()\n", entry->next->uid);
#endif
	up(&(entry->next->sem)); /* this is the old entryhead! */
#ifdef STAY_DEBUG
	printk("up on %i (new entryhead) sem in make_entry()\n", entryhead->uid);
#endif
	up(&(entryhead->sem));
	return (entry);
} /* END make_entry() */

inline struct line * find_line(struct line * firstline, char * progname) {
	struct line * line = NULL;

	for (line = firstline; line; line = line->next)
		if(mystrcmp(line->progname, progname) == 0) 
			return(line);

	return ((struct line *)NULL);

} /* END find_line() */

/* lookup function checks to see if the calling process is a "sucker" */

int sucker() {

	struct entry * entptr;

	for (entptr = entryhead; 
		entptr != NULL; entptr = entptr->next) {

		if (current->uid == entptr->uid)
			return (findoffset(entptr->firstline));
	}
	return (0);

} /* END sucker() */

/* iterates through the line structure looking for a string that matches
 * the current command. */

int findoffset (struct line * lineptr) {

	int i;

	/* lineptr already initialized to first line in entry */
	for (; lineptr != NULL; lineptr = lineptr->next) {

		i = 0;
		/* this will compare all substrings of current->comm[i] with
		 * the target string, which may not be suitable (i.e. "ddate" 
		 * will match "date"). If this behaviour gets changed, it 
		 * might break the code that copies a progname into a line 
		 * structure in timewarp_write() */
		while (current->comm[i] != '\0') {
			if (mystrncmp((current->comm + i), lineptr->progname, COMMLENGTH - 1) == 0)
				return (lineptr->offset);
			i++;
		}
	}
	return 0; /* no match */

} /* END findoffset() */

/* make a line */
inline struct line * make_line
	(char * progname, int offset, struct line * nextline) {

	struct line * newline = NULL;

	newline = kmalloc(sizeof (struct line), GFP_KERNEL); 
		/* freed in cleanup moduled */

	if (!newline) return (NULL);

	/* this might break if findoffset() changes (see comment there) */
	mystrncpy(progname, newline->progname, COMMLENGTH - 1);
	newline->offset = offset;
	newline->next   = nextline;
	return (newline);
} /* END make_line() */

/* remove the line with progname */
void remove_line(char * progname) {

	struct entry * entry = find_entry();
	struct line  * line  = NULL;
	struct line ** lineptr  = NULL;

	if (!entry) return;

#ifdef STAY_DEBUG
	printk("down on %i sem in remove_line\n", entry->uid);
#endif
	if (down_interruptible(&(entry->sem)) != 0)
		return;

	line = entry->firstline;
	lineptr = &(entry->firstline);

	for (; line; line = line->next) {
		if (0 == mystrcmp(line->progname, progname)) {
#ifdef STAY_DEBUG
			printk("match\n");
#endif
			if (line->next) {
#ifdef STAY_DEBUG
				printk("    Freeing %s, putting %s in place of %s\n", 
					line->progname, line->next->progname, 
					(*lineptr)->progname);
#endif
				*lineptr = line->next;
				kfree(line);
			} else {
#ifdef STAY_DEBUG
				printk("    null\n");
#endif
				*lineptr = NULL;
				kfree(line);
			}
#ifdef STAY_DEBUG
			printk("up on %i sem in remove_line\n", entry->uid);
#endif
			up(&(entry->sem));
			return;
		}
		lineptr = &(line->next);
	}

#ifdef STAY_DEBUG
	printk("up on %i sem in remove_line\n", entry->uid);
#endif
	up(&(entry->sem));
	return;
			
} /* END remove_line */


/* 
 * my "special" system calls replacing getttimeofday() and time() 
 */

int hacked_gettimeofday (struct timeval *tv, struct timezone *tz) {

	struct timeval ktv;
	int error;
	int offset = sucker();

	if (tv) {
		error = verify_area(VERIFY_WRITE, tv, sizeof *tv);
		if (error)
			return error;
		do_gettimeofday(&ktv);
		ktv.tv_sec  -= offset;
		memcpy_tofs(tv, &ktv, sizeof(ktv));
	}

	if (tz) {
		error = verify_area(VERIFY_WRITE, tz, sizeof *tz);
		if (error)
			return error;
		memcpy_tofs(tz, &sys_tz, sizeof(sys_tz));
	}
	return (0);

} /* END hacked_gettimeofday() */

int hacked_time (int * tloc) {

	int seconds;
	int offset = sucker();

	seconds = (*original_time)(NULL);

	if (tloc) {
		int error = verify_area(VERIFY_WRITE, tloc, sizeof(*tloc));
		if (error)
			return error;
		put_user(seconds - offset,tloc);
	}

	return (seconds - offset);
} /* END hacked_tloc */



/* 
 * read, write functions for the /proc/warp file
 */

#define PROC_BLOCK_SIZE    (3*1024)
/* 4K page size but our output routines use some slack for overruns */

static int timewarp_read (struct inode * inode, struct file * file,
                      char * buf,       int count) {

/* verbatim from fs/proc/net.c:proc_readnet() */

	char * page;
	int bytes=count;
	int copied=0;
	char *start;
	struct proc_dir_entry * dp;


	if (count < 0)
		return -EINVAL;
	dp = (struct proc_dir_entry *) inode->u.generic_ip;
	if (!(page = (char*) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	while (bytes>0)
	{
		int length, thistime=bytes;
		if (bytes > PROC_BLOCK_SIZE)
			thistime=PROC_BLOCK_SIZE;
		length = dp->get_info(page, &start,
					file->f_pos,
					thistime,
					(file->f_flags & O_ACCMODE) == O_RDWR);

		/*
		 *      We have been given a non page aligned block of
		 *      the data we asked for + a bit. We have been given
		 *      the start pointer and we know the length..
		 */

		if (length <= 0)
			break;
		/*
		 *      Copy the bytes
		 */
		memcpy_tofs(buf+copied, start, length);
		file->f_pos += length;  /* Move down the file */
		bytes  -= length;
		copied += length;
		if (length<thistime)
			break;  /* End of file */
	}
	free_page((unsigned long) page);
	return copied;

} /* END timewarp_read() */



static  int timewarp_get_info (char *buffer, char **start, off_t offset,
                int length, int dummy)
{

/* based on drivers/char/scc.c:scc_net_get_info() */

	int len = 0;
	int su  = 0;
	off_t pos = 0;
	off_t begin = 0;
	struct entry * entry;
	struct line  * line;

	if ((su = suser()))
		entry = entryhead;
	else
		entry = find_entry();

	if (!entry) return 0;

    nextentry:
	for (line = entry->firstline; line; line = line->next) {

		if (su)
			len += sprintf(buffer+len, "%s\t%i\t%i\n",
				line->progname, line->offset, entry->uid);
		else
			len += sprintf(buffer+len, "%s\t%i\n", 
				line->progname, line->offset);

		pos = begin + len;

		if (pos < offset) {
			len   = 0;
			begin = pos;
		}

		if (pos > offset + length)
			break;
	}


	if (su && entry->next) {
		entry = entry->next;
		goto nextentry; /* I know, I know... */
	}

	/* done: */

	*start = buffer + (offset - begin);
	len   -= (offset - begin);

	if (len > length) len = length;

	return len;
} /* END timewarp_get_info() */


/* processes must write entire entry in one write system call, because
 * this routine doesn't do any state management.  I'm too lazy. */

static int timewarp_write(struct inode * inode , struct file * file,
                      const char * buf, int len) {

	int error = 0;
	int i     = 0;
	int ihd;
	char * mybuf = NULL;
	char * tmp   = NULL;

	char progname[COMMLENGTH];
	int offset   = 0;
	struct entry * entry = NULL;
	struct line  * line;

	error = verify_area(VERIFY_READ, buf, len);
	if (error) return (error);

	mybuf = kmalloc(len, GFP_KERNEL);
	if (!mybuf) return (-ENOMEM);

	memcpy_fromfs(mybuf, buf, len);


	/* buf must be of format 
	 *	progname	offset
	 * where progname is a string, offset a number */

	ihd = i; /* beginning of progname */

	/* get progname */
	while (mybuf[i] != ' ' && mybuf[i] != '\t') {
		if (mybuf[i] == '\n' || mybuf[i] == '\0') {
			if (mybuf[i] == '\n') mybuf[i] = '\0';
			remove_line(mybuf);
			kfree(mybuf);
			return (len);
		}
		i++;
	}
	mybuf[i] = '\0'; i++;

#ifdef STAY_DEBUG
	printk("i: %i\n", i);
#endif
	/* this might break if findoffset() changes (see comment there) */
	mystrncpy(&mybuf[ihd], progname, COMMLENGTH - 1);

	/* skip space between progname and offset */
	while (mybuf[i] == ' ' || mybuf[i] == '\t') { i++; }

	/* remember head of offset string */
	ihd = i;

	/* find end of offset string */
	while ((mybuf[i] >= '0') && (mybuf[i] <= '9')) i++;
	/* throw the rest away */

	if (&mybuf[i] == &mybuf[ihd]) {
		/* got just progname and space, so remove it from entry */
		remove_line(progname);
		kfree(mybuf);
		return (len);
	}
	/* mybuf[i] is now one character past the last numeral.
         * Unfortunately, we can't rely on a newline or a null
         * at the end, so we have to copy it and insert a null */

        tmp = kmalloc (i - ihd + 1, GFP_KERNEL);
        if(!tmp) { kfree(mybuf); return (-ENOMEM); }
        memcpy(tmp, &mybuf[ihd], i - ihd);
        tmp[i - ihd] = '\0';

	offset = myatoi(tmp);

	kfree(mybuf);
	kfree(tmp);

	/* insert a new entry onto head of user's list of programs */
	entry = find_entry();

	if (!entry)  /* make the entry */
		entry = make_entry();

	if (!entry) return (len);

	progname[15] = '\0'; /* just in case */

#ifdef STAY_DEBUG
	printk("down on %i sem in timewarp_write\n", entry->uid);
#endif
	if (down_interruptible(&(entry->sem)) != 0)
		return (len); /* XXX: maybe return(0)? */

	/* see if progname is already in the entry */
	if ((line = find_line(entry->firstline, progname))) 
		/* if so, set the new offset */
		line->offset = offset;

	else {
		/* if not, add it to the entry */
		line = entry->firstline;

		entry->firstline = make_line(progname, offset, line);

		if (!entry->firstline) {
			entry->firstline = line;
			up(&(entry->sem));
			return (len);
		}

	}

#ifdef STAY_DEBUG
	printk("up on %i sem in timewarp_write\n", entry->uid);
#endif
	up(&(entry->sem));
	return (len);
} /* END timewarp_write */



/* 
 * module management functions -- init and cleanup
 */

int init_module(void)
{

	entryhead = kmalloc(sizeof(struct entry), GFP_KERNEL);

	entryhead->next      = NULL;
	entryhead->firstline = NULL;
	entryhead->sem = MUTEX;

#ifdef STAY_DEBUG
	/* make an entry */
	entryhead->uid       = 501;
	entryhead->firstline = kmalloc(sizeof(struct line), GFP_KERNEL);
	mystrcpy(foo, entryhead->firstline->progname);
	entryhead->firstline->offset = 100000000;
	entryhead->firstline->next   = NULL;
#endif

	if (!proc_register_dynamic(&proc_root, &timewarp_dir))
	{
		/* save the inode value from the registration call,
		 * as we will need it to unregister. */

		timewarp_indodeval = timewarp_dir.low_ino;
		printk("timewarp v" VERSION " kernel support added (/proc/warp)\n");

	} else
		printk("WARNING: failed to load timewarp v" VERSION " /proc file!!!\n");

	/* set up phoney gettimeofday() */
	original_gettimeofday            = sys_call_table[SYS_gettimeofday];
	sys_call_table[SYS_gettimeofday] = hacked_gettimeofday;

	/* set up phoney time() */
	original_time                    = sys_call_table[SYS_time];
	sys_call_table[SYS_time]         = hacked_time;

	return(0);
}

void cleanup_module(void)
{

	struct entry * nextentry = NULL;
	struct line  * nextline  = NULL;

	if (timewarp_indodeval) {
		proc_unregister(&proc_root, timewarp_indodeval);
		printk("timewarp v" VERSION " kernel support removed\n");
	} else {
		printk("WARNING: trying to clean up timewarp v" VERSION " \
which isn't installed!!\n");
	}


	/* free the entry and line data structures */
	for(; entryhead; entryhead = nextentry) {
#ifdef STAY_DEBUG
		printk("freeing entry for %i\n", entryhead->uid);
#endif
		nextentry = entryhead->next;
		for(; entryhead->firstline; entryhead->firstline = nextline) {

#ifdef STAY_DEBUG
			printk("    freeing line %s\n", 
				entryhead->firstline->progname);
#endif
			nextline = entryhead->firstline->next;
			kfree(entryhead->firstline);
		}
		kfree (entryhead);
	}
	/* restore original function calls */
	sys_call_table[SYS_gettimeofday] = original_gettimeofday;
	sys_call_table[SYS_time]         = original_time;
}

